# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import empty, matmul, diag_embed
from torch.nn import Embedding, Parameter
from torch.nn.functional import normalize
from torch.nn.init import xavier_uniform_

from torchkge.models import Model
from torchkge.utils import get_rank, get_mask, get_rolling_matrix, get_true_targets
from torchkge.exceptions import WrongDimensionError


class RESCALModel(Model):
    """Implementation of RESCAL model detailed in 2011 paper by Nickel et al..\
    In the original paper, optimization is done using Alternating Least Squares (ALS). Here we use\
    iterative gradient descent optimization. This class inherits from the
    :class:`torchkge.models.interfaces.Model` interface. It then has its attributes as well.

    References
    ----------
    * Maximilian Nickel, Volker Tresp, and Hans-Peter Kriegel.
      `A Three-way Model for Collective Learning on Multi-relational Data.
      <https://dl.acm.org/citation.cfm?id=3104584>`_
      In Proceedings of the 28th International Conference on Machine Learning, 2011.


    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    entity_embeddings: torch.nn.Embedding, shape: (number_entities, ent_emb_dim)
        Contains the embeddings of the entities. It is initialized with Xavier uniform and then\
         normalized.
    relation_matrices: torch Parameter, shape: (number_relations, ent_emb_dim, ent_emb_dim)
        Contains the matrices of the relations. It is initialized with Xavier uniform.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__(ent_emb_dim, n_entities, n_relations)

        # initialize embedding objects
        self.entity_embeddings = Embedding(self.number_entities, self.ent_emb_dim)
        self.relation_matrices = Parameter(xavier_uniform_(empty(size=(self.number_relations,
                                                                       self.ent_emb_dim,
                                                                       self.ent_emb_dim))),
                                           requires_grad=True)

        # fill the embedding weights with Xavier initialized values
        self.entity_embeddings.weight = Parameter(xavier_uniform_(
            empty(size=(self.number_entities, self.ent_emb_dim))),
            requires_grad=True)

        # normalize the embeddings
        self.normalize_parameters()

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        # recover entities embeddings
        heads_embeddings = normalize(self.entity_embeddings(heads_idx), p=2, dim=1)
        tails_embeddings = normalize(self.entity_embeddings(tails_idx), p=2, dim=1)

        # recover relation matrices
        relation_matrices = self.relation_matrices[rels_idx]

        product = self.compute_product(heads_embeddings, tails_embeddings, relation_matrices)
        return product.view(heads_idx.shape[0])

    def normalize_parameters(self):
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)

    def compute_product(self, heads, tails, rel_mat):
        """Compute the matrix product h^tRt with proper reshapes. It can do the batch matrix
        product both in the forward pass and in the evaluation pass with one matrix containing
        all candidates.

        Parameters
        ----------
        heads: torch Tensor, shape: (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities or candidates.
        tails: `torch.Tensor`, shape: (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities or canditates.
        rel_mat: `torch.Tensor`, shape: (b_size, self.ent_emb_dim, self.ent_emb_dim), dtype: `torch.float`
            Tensor containing relation matrices for current relations.

        Returns
        -------
        product: `torch.Tensor`, shape: (b_size) or (b_size, self.number_entities), dtype: `torch.float`
            Tensor containing the matrix products h.W.t^t for each sample of the batch.

        """
        b_size = len(heads)

        if len(heads.shape) == 2 and len(tails.shape) == 2:
            heads = heads.view(b_size, 1, self.ent_emb_dim)
            tails = tails.view(b_size, self.ent_emb_dim, 1)
            return matmul(matmul(heads, rel_mat), tails).view(b_size)

        elif len(heads.shape) == 2 and len(tails.shape) == 3:
            heads = heads.view(b_size, 1, self.ent_emb_dim)
            tails = tails.transpose(1, 2)
        else:
            tails = tails.view(b_size, self.ent_emb_dim, 1)

        return matmul(matmul(heads, rel_mat), tails).view(b_size, -1)

    def get_head_tail_candidates(self, h_idx, t_idx):
        b_size = h_idx.shape[0]

        candidates = self.entity_embeddings.weight.data
        candidates = candidates.view(1, self.number_entities, self.ent_emb_dim)
        candidates = candidates.expand(b_size, self.number_entities, self.ent_emb_dim)

        h_emb = self.entity_embeddings(h_idx)
        t_emb = self.entity_embeddings(t_idx)

        return h_emb, t_emb, candidates

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        h_emb: `torch.Tensor`, shape: (b_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities.
        t_emb: `torch.Tensor`, shape: (b_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities.
        r_mat: `torch.Tensor`, shape: (b_size, ent_emb_dim, ent_emb_dim), dtype: `torch.float`
            Tensor containing matrices of current relations.
        candidates: `torch.Tensor`, shape: (b_size, number_entities, ent_emb_dim), dtype: `torch.float`
            Tensor containing all entities as candidates for each sample of the batch.

        """
        h_emb, t_emb, candidates = self.get_head_tail_candidates(h_idx, t_idx)
        r_mat = self.relation_matrices[r_idx]

        return h_emb, t_emb, candidates, r_mat

    def compute_ranks(self, e_emb, candidates, r_mat, e_idx, r_idx, true_idx, dictionary, heads=1):
        """Compute the ranks and the filtered ranks of true entities when doing link prediction.

        Parameters
        ----------
        e_emb: torch tensor, shape: (batch_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current embeddings of entities.
        candidates: torch tensor, shape: (b_size, number_entities, ent_emb_dim), dtype: `torch.float`
            Tensor containing projected embeddings of all entities.
        r_mat: `torch.Tensor`, shape: (b_size, ent_emb_dim, ent_emb_dim), dtype: `torch.float`
            Tensor containing current matrices of relations.
        e_idx: torch tensor, shape: (batch_size), dtype: `torch.long`
            Tensor containing the indices of entities.
        r_idx: torch tensor, shape: (batch_size), dtype: `torch.long`
            Tensor containing the indices of relations.
        true_idx: torch tensor, shape: (batch_size), dtype: `torch.long`
            Tensor containing the true entity for each sample.
        dictionary: default dict
            Dictionary of keys (int, int) and values list of ints giving all possible entities for\
            the (entity, relation) pair.
        heads: integer
            1 ou -1 (must be 1 if entities are heads and -1 if entities are tails). We test\
             dissimilarity_type between heads * entities + relations and heads * targets.


        Returns
        -------
        rank_true_entities: torch Tensor, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true entities when ranking any entity based on\
            estimation of 1 or 0.
        filtered_rank_true_entities: torch Tensor, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true entities when ranking only true false entities\
            based on estimation of 1 or 0.

        """
        current_batch_size, _ = e_emb.shape

        if heads == 1:
            scores = self.compute_product(e_emb, candidates, r_mat)
        else:
            scores = self.compute_product(candidates, e_emb, r_mat)

        # filter out the true negative samples by assigning negative score
        filt_scores = scores.clone()
        for i in range(current_batch_size):
            true_targets = get_true_targets(dictionary, e_idx, r_idx, true_idx, i)
            if true_targets is None:
                continue
            filt_scores[i][true_targets] = float(-1)

        # from dissimilarities, extract the rank of the true entity.
        rank_true_entities = get_rank(scores, true_idx)
        filtered_rank_true_entities = get_rank(filt_scores, true_idx)

        return rank_true_entities, filtered_rank_true_entities


class DistMultModel(RESCALModel):
    """Implementation of DistMult model detailed in 2014 paper by Yang et al.. This class inherits from the
    :class:`torchkge.models.BilinearModels.RESCALModel` class interpreted as an interface.
    It then has its attributes as well.

    References
    ----------
    * Bishan Yang, Wen-tau Yih, Xiaodong He, Jianfeng Gao, and Li Deng.
      `Embedding Entities and Relations for Learning and Inference in Knowledge Bases.
      <https://arxiv.org/abs/1412.6575>`_
      arXiv :1412.6575 [cs], December 2014.


    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    relation_vectors: torch Parameter, shape: (number_relations, ent_emb_dim)
        Contains the vectors to build diagonal matrices of the relations. It is initialized with
        Xavier uniform.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__(ent_emb_dim, n_entities, n_relations)

        del self.relation_matrices
        self.relation_vectors = Parameter(
            xavier_uniform_(empty(size=(self.number_relations, self.ent_emb_dim))),
            requires_grad=True)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        # recover entities embeddings
        heads_embeddings = normalize(self.entity_embeddings(heads_idx), p=2, dim=1)
        tails_embeddings = normalize(self.entity_embeddings(tails_idx), p=2, dim=1)

        # recover relation matrices
        relation_matrices = diag_embed(self.relation_vectors[rels_idx])

        product = self.compute_product(heads_embeddings, tails_embeddings, relation_matrices)

        return product.view(heads_idx.shape[0])

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        h_emb: `torch.Tensor`, shape: (b_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities.
        t_emb: `torch.Tensor`, shape: (b_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities.
        r_mat: `torch.Tensor`, shape: (b_size, ent_emb_dim, ent_emb_dim), dtype: `torch.float`
            Tensor containing matrices of current relations.
        candidates: `torch.Tensor`, shape: (b_size, number_entities, ent_emb_dim), dtype: `torch.float`
            Tensor containing all entities as candidates for each sample of the batch.

        """
        h_emb, t_emb, candidates = self.get_head_tail_candidates(h_idx, t_idx)
        r_mat = diag_embed(self.relation_vectors[r_idx])

        return h_emb, t_emb, candidates, r_mat


class HolEModel(RESCALModel):
    """Implementation of HolE model detailed in 2015 paper by Nickel et al.. This class inherits from the
    :class:`torchkge.models.BilinearModels.RESCALModel` class interpreted as an interface.
    It then has its attributes as well.

    References
    ----------
    * Maximilian Nickel, Lorenzo Rosasco, and Tomaso Poggio.
      `Holographic Embeddings of Knowledge Graphs.
      <https://arxiv.org/abs/1510.04935>`_
      arXiv :1510.04935 [cs, stat], October 2015.

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities
    relation_vectors: torch Parameter, shape: (number_relations, ent_emb_dim)
        Contains the vectors to build circular matrices of the relations. It is initialized
        with Xavier uniform.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__(ent_emb_dim, n_entities, n_relations)

        del self.relation_matrices
        self.relation_vectors = Parameter(
            xavier_uniform_(empty(size=(self.number_relations, self.ent_emb_dim))), requires_grad=True)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        # recover entities embeddings
        heads_embeddings = normalize(self.entity_embeddings(heads_idx), p=2, dim=1)
        tails_embeddings = normalize(self.entity_embeddings(tails_idx), p=2, dim=1)

        # recover relation matrices
        relation_matrices = get_rolling_matrix(self.relation_vectors[rels_idx])

        product = self.compute_product(heads_embeddings, tails_embeddings, relation_matrices)

        return product.view(heads_idx.shape[0])

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size,), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        h_emb: `torch.Tensor`, shape: (b_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities.
        t_emb: `torch.Tensor`, shape: (b_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities.
        r_mat: `torch.Tensor`, shape: (b_size, ent_emb_dim, ent_emb_dim), dtype: `torch.float`
            Tensor containing matrices of current relations.
        candidates: `torch.Tensor`, shape: (b_size, number_entities, ent_emb_dim), dtype: `torch.float`
            Tensor containing all entities as candidates for each sample of the batch.

        """
        h_emb, t_emb, candidates = self.get_head_tail_candidates(h_idx, t_idx)
        r_mat = get_rolling_matrix(self.relation_vectors[r_idx])

        return h_emb, t_emb, candidates, r_mat

    def normalize_parameters(self):
        pass


class ComplExModel(DistMultModel):
    """Implementation of ComplEx model detailed in 2016 paper by Trouillon et al.. This class inherits from the
    :class:`torchkge.models.BilinearModels.DistMultModel` class interpreted as an interface.
    It then has its attributes as well.

    References
    ----------
    * Théo Trouillon, Johannes Welbl, Sebastian Riedel, Éric Gaussier, and Guillaume Bouchard.
      `Complex Embeddings for Simple Link Prediction.
      <https://arxiv.org/abs/1606.06357>`_
      arXiv :1606.06357 [cs, stat], June 2016.

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    smaller_dim: int
        Number of 2x2 matrices on the diagonals of relation-specific matrices.
    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        try:
            assert ent_emb_dim % 2 == 0
        except AssertionError:
            raise WrongDimensionError('Embedding dimension should be pair.')

        super().__init__(ent_emb_dim, n_entities, n_relations)

        self.smaller_dim = int(self.ent_emb_dim / 2)

        self.real_mask = get_mask(self.ent_emb_dim, 0, self.smaller_dim)
        self.im_mask = get_mask(self.ent_emb_dim, self.smaller_dim, self.ent_emb_dim)

    def compute_product(self, heads, tails, rel_mat):
        """Compute the matrix product h^tRt with proper reshapes. It can do the batch matrix
        product both in the forward pass and in the evaluation pass with one matrix containing
        all candidates.

        Parameters
        ----------
        heads: torch Tensor, shape: (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities or candidates.
        tails: `torch.Tensor`, shape: (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities or canditates.
        rel_mat: `torch.Tensor`, shape: (b_size, self.ent_emb_dim, self.ent_emb_dim), dtype: `torch.float`
            Tensor containing relation matrices for current relations.

        Returns
        -------
        product: `torch.Tensor`, shape: (b_size, 1) or (b_size, self.number_entities), dtype: `torch.float`
            Tensor containing the matrix products h^t.W.t for each sample of the batch.

        """
        b_size = len(heads)
        r_re = rel_mat[:, self.real_mask][:, :, self.real_mask]
        r_im = rel_mat[:, self.im_mask][:, :, self.im_mask]

        if len(heads.shape) == 2 and len(tails.shape) == 2:
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        elif len(heads.shape) == 2 and len(tails.shape) == 3:
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_re = tails[:, :, self.real_mask].transpose(2, 1)
            t_im = tails[:, :, self.im_mask].transpose(2, 1)

        else:
            h_re = heads[:, :, self.real_mask]
            h_im = heads[:, :, self.im_mask]

            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        re = matmul(h_re, matmul(r_re, t_re) + matmul(r_im, t_im)).view(b_size, -1)
        im = matmul(h_im, matmul(r_re, t_im) - matmul(r_im, t_re)).view(b_size, -1)

        return re + im

    def normalize_parameters(self):
        pass


class AnalogyModel(DistMultModel):
    """Implementation of ANALOGY model detailed in 2017 paper by Liu et al..\
    According to their remark in the implementation details, the number of scalars on\
    the diagonal of each relation-specific matrix is by default set to be half the embedding\
    dimension. This class inherits from the
    :class:`torchkge.models.BilinearModels.DistMultModel` class interpreted as an interface.
    It then has its attributes as well.

    References
    ----------
    * Hanxiao Liu, Yuexin Wu, and Yiming Yang.
      `Analogical Inference for Multi-Relational Embeddings.
      <https://arxiv.org/abs/1705.02426>`_
      arXiv :1705.02426 [cs], May 2017.

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of embedding space.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.
    scalar_share: float
        Share of the diagonal elements of the relation-specific matrices to be scalars. By default\
        it is set to half according to the original paper.

    Attributes
    ----------
    number_scalars: int
        Number of diagonal elements of the relation-specific matrices to be scalars. By default\
        it is set to half according to the original paper.
    smaller_dim: int
        Number of 2x2 matrices on the diagonals of relation-specific matrices.
    """

    def __init__(self, ent_emb_dim, n_entities, n_relations, scalar_share=0.5):
        try:
            assert ent_emb_dim % 2 == 0
        except AssertionError:
            raise WrongDimensionError('Embedding dimension should be pair.')

        super().__init__(ent_emb_dim, n_entities, n_relations)

        self.number_scalars = int(self.ent_emb_dim * scalar_share)

        if (self.ent_emb_dim - self.number_scalars) % 2 == 1:
            # the diagonal blocks are 2x2 so this dimension needs to be pair
            self.number_scalars -= 1
        self.smaller_dim = int((self.ent_emb_dim - self.number_scalars) / 2)

        self.scalar_mask = get_mask(self.ent_emb_dim, 0, self.number_scalars)
        self.real_mask = get_mask(self.ent_emb_dim, self.number_scalars,
                                  self.number_scalars + self.smaller_dim)
        self.im_mask = get_mask(self.ent_emb_dim, self.number_scalars + self.smaller_dim,
                                self.ent_emb_dim)

        assert (self.real_mask.sum() == self.im_mask.sum() == self.smaller_dim)

    def compute_product(self, heads, tails, rel_mat):
        """Compute the matrix product h^tRt with proper reshapes. It can do the batch matrix
        product both in the forward pass and in the evaluation pass with one matrix containing
        all candidates.

        Parameters
        ----------
        heads: torch Tensor, shape: (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities or candidates.
        tails: `torch.Tensor`, shape: (b_size, self.ent_emb_dim) or (b_size, self.number_entities,\
        self.ent_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities or canditates.
        rel_mat: `torch.Tensor`, shape: (b_size, self.ent_emb_dim, self.ent_emb_dim), dtype: `torch.float`
            Tensor containing relation matrices for current relations.

        Returns
        -------
        product: `torch.Tensor`, shape: (b_size, 1) or (b_size, self.number_entities), dtype: `torch.float`
            Tensor containing the matrix products h^t.W.t for each sample of the batch.

        """
        b_size = len(heads)
        r_scalar = rel_mat[:, self.scalar_mask][:, :, self.scalar_mask]
        r_re = rel_mat[:, self.real_mask][:, :, self.real_mask]
        r_im = rel_mat[:, self.im_mask][:, :, self.im_mask]

        if len(heads.shape) == 2 and len(tails.shape) == 2:
            h_scalar = heads[:, self.scalar_mask].view(b_size, 1, self.number_scalars)
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_scalar = tails[:, self.scalar_mask].view(b_size, self.number_scalars, 1)
            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        elif len(heads.shape) == 2 and len(tails.shape) == 3:
            h_scalar = heads[:, self.scalar_mask].view(b_size, 1, self.number_scalars)
            h_re = heads[:, self.real_mask].view(b_size, 1, self.smaller_dim)
            h_im = heads[:, self.im_mask].view(b_size, 1, self.smaller_dim)

            t_scalar = tails[:, :, self.scalar_mask].transpose(2, 1)
            t_re = tails[:, :, self.real_mask].transpose(2, 1)
            t_im = tails[:, :, self.im_mask].transpose(2, 1)

        else:
            h_scalar = heads[:, :, self.scalar_mask]
            h_re = heads[:, :, self.real_mask]
            h_im = heads[:, :, self.im_mask]

            t_scalar = tails[:, self.scalar_mask].view(b_size, self.number_scalars, 1)
            t_re = tails[:, self.real_mask].view(b_size, self.smaller_dim, 1)
            t_im = tails[:, self.im_mask].view(b_size, self.smaller_dim, 1)

        scalar = matmul(matmul(h_scalar, r_scalar), t_scalar).view(b_size, -1)
        re = matmul(h_re, matmul(r_re, t_re) + matmul(r_im, t_im)).view(b_size, -1)
        im = matmul(h_im, matmul(r_re, t_im) - matmul(r_im, t_re)).view(b_size, -1)

        return scalar + re + im

    def normalize_parameters(self):
        """According to original paper, no normalization should be done on the parameters.

        """
        pass



from .dissimilarities import l1_dissimilarity, l2_dissimilarity
from .dissimilarities import l1_torus_dissimilarity, l2_torus_dissimilarity, el2_torus_dissimilarity
from .losses import MarginLoss, LogisticLoss

from .models_utils import init_embedding, get_true_targets
from .negative_sampling import get_possible_heads_tails
from .operations import get_rank, get_mask, get_rolling_matrix
from .preprocessing import get_dictionaries, get_bernoulli_probs



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import cat, zeros


def get_mask(length, start, end):
    """Create a mask of length `length` filled with 0s except between indices `start` (included)\
    and `end` (excluded).

    Parameters
    ----------
    length: int
        Length of the mask to be created.
    start: int
        First index (included) where the mask will be filled with 0s.
    end: int
        Last index (excluded) where the mask will be filled with 0s.

    Returns
    -------
    mask: `torch.Tensor`, shape: (length), dtype: `torch.bool`
        Mask of length `length` filled with 0s except between indices `start` (included)\
        and `end` (excluded).
    """
    mask = zeros(length)
    mask[[i for i in range(start, end)]] = 1
    return mask.bool()


def get_rank(data, true, low_values=False):
    """Computes how many entities have higher (or lower) value in data than the one at index true[i]

    Parameters
    ----------
    data: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dimensions)
        Scores for each entity.
    true: `torch.Tensor`, dtype: `torch.int`, shape: (n_facts)
        true[i] is the index of the true entity for test i of the batch.
    low_values: bool, optional (default=False)
        if True, best rank is the lowest score else it is the highest.

    Returns
    -------
    ranks: `torch.Tensor`, dtype: `torch.int`, shape: (n_facts)
        ranks[i] is the number of entities which have better scores in data than the one and index true[i]
    """
    true_data = data.gather(1, true.long().view(-1, 1))

    if low_values:
        return (data <= true_data).sum(dim=1)
    else:
        return (data >= true_data).sum(dim=1)


def get_rolling_matrix(x):
    """Build a rolling matrix.

    Parameters
    ----------
    x: `torch.Tensor`, shape: (b_size, dim)

    Returns
    -------
    mat: `torch.Tensor`, shape: (b_size, dim, dim)
        Rolling matrix such that mat[i,j] = x[i - j mod(dim)]
    """
    b_size, dim = x.shape
    x = x.view(b_size, 1, dim)
    return cat([x.roll(i, dims=2) for i in range(dim)], dim=1)



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import abs, min, sqrt, cos
from math import pi


def l1_dissimilarity(a, b):
    """
    Parameters
    ----------
    a: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)
    b: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)

    Returns
    -------
    dissimilarity_type: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts)
        Tensor of the row_wise L1 distance.

    """
    return (a-b).norm(p=1, dim=1)


def l2_dissimilarity(a, b):
    """
    Parameters
    ----------
    a: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)
    b: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)

    Returns
    -------
    dissimilarity_type: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts)
        Tensor of the row_wise squared L2 distance.
    """
    return (a-b).norm(p=2, dim=1)**2


def l1_torus_dissimilarity(a, b):
    """See paper by Ebisu et al. for details about the definition of this dissimilarity_type function.

    Parameters
    ----------
    a: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)
    b: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)

    Returns
    -------
    dissimilarity_type: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts)
        Tensor of the row_wise squared L1 distance on the Torus.
    """
    a, b = a.frac(), b.frac()
    return min(abs(a-b), 1 - abs(a-b)).sum(dim=1)


def l2_torus_dissimilarity(a, b):
    """See paper by Ebisu et al. for details about the definition of this dissimilarity_type function.

    Parameters
    ----------
    a: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)
    b: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)

    Returns
    -------
    dissimilarity_type: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts)
        Tensor of the row_wise squared L2 distance on the Torus.
    """
    a, b = a.frac(), b.frac()
    return sqrt(min((a - b)**2, 1 - (a - b)**2).sum(dim=1))


def el2_torus_dissimilarity(a, b):
    """See paper by Ebisu et al. for details about the definition of this dissimilarity_type function.

    Parameters
    ----------
    a: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)
    b: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts, dim)

    Returns
    -------
    dissimilarity_type: `torch.Tensor`, dtype: `torch.float`, shape: (n_facts)
        Tensor of the row_wise squared L2 distance on the Torus derived from the distance in :math:`\\mathbb{C}^n`.

    """
    tmp = min(a - b, 1 - (a-b))
    tmp = 2 * (1 - cos(2 * pi * tmp))
    return sqrt(tmp.sum(dim=1))



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import arange
from torch.nn import Module

from torchkge.utils import init_embedding, l1_dissimilarity, l2_dissimilarity
from torchkge.utils import get_rank, get_true_targets


class Model(Module):
    """Model interface to be used by any other class implementing a knowledge graph embedding model.

    Parameters
    ----------
    ent_emb_dim: int
        Embedding dimension of the entities.
    n_entities: int
        Number of entities to be embedded.
    n_relations: int
        Number of relations to be embedded.

    Attributes
    ----------
    ent_emb_dim: int
        Embedding dimension of the entities.
    number_entities: int
        Number of entities to be embedded.
    number_relations: int
        Number of relations to be embedded.

    """
    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__()
        self.ent_emb_dim = ent_emb_dim
        self.number_entities = n_entities
        self.number_relations = n_relations

    def forward(self, heads, tails, negative_heads, negative_tails, relations):
        """Forward pass on the current batch.

        Parameters
        ----------
        heads: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        negative_heads: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's negatively sampled heads.
        negative_tails: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's negatively sampled tails.
        relations: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        golden_triplets: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Scoring function evaluated on true triples.
        negative_triplets: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Scoring function evaluated on negatively sampled triples.

        """
        return self.scoring_function(heads, tails, relations), \
            self.scoring_function(negative_heads, negative_tails, relations)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Score function: opposite of dissimilarities between h+r and t.

        """
        raise NotImplementedError

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        raise NotImplementedError

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: `torch.Tensor`, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current relations embeddings.

        """
        raise NotImplementedError

    def compute_ranks(self, proj_e_emb, proj_candidates, r_emb, e_idx, r_idx, true_idx, dictionary, heads=1):
        """Compute the ranks and the filtered ranks of true entities when doing link prediction.

        Parameters
        ----------
        proj_e_emb: `torch.Tensor`, shape: (batch_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current projected embeddings of entities.
        proj_candidates: `torch.Tensor`, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing projected embeddings of all entities.
        r_emb: `torch.Tensor`, shape: (batch_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing current embeddings of relations.
        e_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
            Tensor containing the indices of entities.
        r_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
            Tensor containing the indices of relations.
        true_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
            Tensor containing the true entity for each sample.
        dictionary: default dict
            Dictionary of keys (int, int) and values list of ints giving all possible entities for
            the (entity, relation) pair.
        heads: integer
            1 ou -1 (must be 1 if entities are heads and -1 if entities are tails). \
            We test dissimilarity_type between heads * entities + relations and heads * targets.


        Returns
        -------
        rank_true_entities: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true entities when ranking any entity based on \
            computation of d(hear+relation, tail).
        filtered_rank_true_entities: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true entities when ranking only true false entities \
            based on computation of d(hear+relation, tail).

        """
        raise NotImplementedError

    def evaluate_candidates(self, h_idx, t_idx, r_idx, kg):
        """Compute the head and tail ranks and filtered ranks of the current batch.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current relations.
        kg: `torchkge.data.KnowledgeGraph.KnowledgeGraph`
            Knowledge graph on which the model was trained.

        Returns
        -------
        rank_true_tails: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true tails when ranking any entity based on \
            computation of d(hear+relation, tail).
        filt_rank_true_tails: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true tails when ranking only true false entities \
            based on computation of d(hear+relation, tail).
        rank_true_heads: Tensor containing the rank of the true heads when ranking any entity based on \
            computation of d(hear+relation, tail).
        filt_rank_true_heads: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true heads when ranking only true false entities \
            based on computation of d(hear+relation, tail).

        """
        proj_h_emb, proj_t_emb, candidates, r_emb = self.evaluation_helper(h_idx, t_idx, r_idx)

        # evaluation_helper both ways (head, rel) -> tail and (rel, tail) -> head
        rank_true_tails, filt_rank_true_tails = self.compute_ranks(proj_h_emb,
                                                                   candidates,
                                                                   r_emb, h_idx, r_idx,
                                                                   t_idx,
                                                                   kg.dict_of_tails,
                                                                   heads=1)
        rank_true_heads, filt_rank_true_heads = self.compute_ranks(proj_t_emb,
                                                                   candidates,
                                                                   r_emb, t_idx, r_idx,
                                                                   h_idx,
                                                                   kg.dict_of_heads,
                                                                   heads=-1)

        return rank_true_tails, filt_rank_true_tails, rank_true_heads, filt_rank_true_heads


class TranslationalModel(Model):
    """Model interface to be used by any other class implementing a translational knowledge graph embedding model.
    This interface inherits from the interface :class:`torchkge.models.interfaces.Model`.

    Parameters
    ----------
    ent_emb_dim: int
        Embedding dimension of the entities.
    n_entities: int
        Number of entities to be embedded.
    n_relations: int
        Number of relations to be embedded.
    dissimilarity_type: string
        Name of the dissimilarity function to be used.

    Attributes
    ----------
    entity_embeddings: `torch.nn.Embedding`
        Embedding object containing the embeddings of the entities.
    dissimilarity: function
        Dissimilarity function defined in `torchkge.utils.dissimilarities`.

    """
    def __init__(self, ent_emb_dim, n_entities, n_relations, dissimilarity_type):
        super().__init__(ent_emb_dim, n_entities, n_relations)

        self.entity_embeddings = init_embedding(self.number_entities, self.ent_emb_dim)

        assert dissimilarity_type in ['L1', 'L2', None]
        if dissimilarity_type == 'L1':
            self.dissimilarity = l1_dissimilarity
        elif dissimilarity_type == 'L2':
            self.dissimilarity = l2_dissimilarity
        else:
            self.dissimilarity = None

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        raise NotImplementedError

    def normalize_parameters(self):
        raise NotImplementedError

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        raise NotImplementedError

    def recover_project_normalize(self, ent_idx, rel_idx, normalize_):
        raise NotImplementedError

    def compute_ranks(self, proj_e_emb, proj_candidates, r_emb, e_idx, r_idx, true_idx, dictionary, heads=1):
        """Compute the ranks and the filtered ranks of true entities when doing link prediction.

        Parameters
        ----------
        proj_e_emb: `torch.Tensor`, shape: (batch_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current projected embeddings of entities.
        proj_candidates: `torch.Tensor`, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing projected embeddings of all entities.
        r_emb: `torch.Tensor`, shape: (batch_size, ent_emb_dim), dtype: `torch.float`
            Tensor containing current embeddings of relations.
        e_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
            Tensor containing the indices of entities.
        r_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
            Tensor containing the indices of relations.
        true_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
            Tensor containing the true entity for each sample.
        dictionary: default dict
            Dictionary of keys (int, int) and values list of ints giving all possible entities for
            the (entity, relation) pair.
        heads: integer
            1 ou -1 (must be 1 if entities are heads and -1 if entities are tails). \
            We test dissimilarity_type between heads * entities + relations and heads * targets.


        Returns
        -------
        rank_true_entities: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true entities when ranking any entity based on \
            computation of d(hear+relation, tail).
        filtered_rank_true_entities: `torch.Tensor`, shape: (b_size), dtype: `torch.int`
            Tensor containing the rank of the true entities when ranking only true false entities \
            based on computation of d(hear+relation, tail).

        """
        current_batch_size, embedding_dimension = proj_e_emb.shape

        # tmp_sum is either heads + r_emb or r_emb - tails (expand does not use extra memory)
        tmp_sum = (heads * proj_e_emb + r_emb).view((current_batch_size, embedding_dimension, 1))
        tmp_sum = tmp_sum.expand((current_batch_size, embedding_dimension, self.number_entities))

        # compute either dissimilarity_type(heads + relation, proj_candidates) or
        # dissimilarity_type(-proj_candidates, relation - tails)
        dissimilarities = self.dissimilarity(tmp_sum, heads * proj_candidates)

        # filter out the true negative samples by assigning infinite dissimilarity_type
        filt_dissimilarities = dissimilarities.clone()
        for i in range(current_batch_size):
            true_targets = get_true_targets(dictionary, e_idx, r_idx, true_idx, i)
            if true_targets is None:
                continue
            filt_dissimilarities[i][true_targets] = float('Inf')

        # from dissimilarities, extract the rank of the true entity.
        rank_true_entities = get_rank(-dissimilarities, true_idx)
        filtered_rank_true_entities = get_rank(-filt_dissimilarities, true_idx)

        return rank_true_entities, filtered_rank_true_entities

    def recover_candidates(self, h_idx, b_size):
        """Prepare candidates for link prediction evaluation.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        b_size: int
            Batch size.

        Returns
        -------
        candidates: `torch.Tensor`, shape: (b_size, ent_emb_dim, number_entities), dtype: `torch.float`
            Tensor containing replications of all entities embeddings as many times as the batch size.

        """
        all_idx = arange(0, self.number_entities).long()
        if h_idx.is_cuda:
            all_idx = all_idx.cuda()
        candidates = self.entity_embeddings(all_idx).transpose(0, 1)
        candidates = candidates.view((1,
                                      self.ent_emb_dim,
                                      self.number_entities)).expand((b_size,
                                                                     self.ent_emb_dim,
                                                                     self.number_entities))
        return candidates

    @staticmethod
    def projection_helper(h_idx, t_idx, b_size, candidates, rel_emb_dim):
        mask = h_idx.view(b_size, 1, 1).expand(b_size, rel_emb_dim, 1)
        proj_h_emb = candidates.gather(dim=2, index=mask).view(b_size, rel_emb_dim)

        mask = t_idx.view(b_size, 1, 1).expand(b_size, rel_emb_dim, 1)
        proj_t_emb = candidates.gather(dim=2, index=mask).view(b_size, rel_emb_dim)

        return proj_h_emb, proj_t_emb



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import empty, matmul, eye, arange, tensor
from torch.nn import Parameter
from torch.nn.functional import normalize
from torch.nn.init import xavier_uniform_
from torch.cuda import empty_cache

from torchkge.models import TranslationalModel
from torchkge.utils import init_embedding
from torchkge.utils import l1_torus_dissimilarity, l2_torus_dissimilarity, el2_torus_dissimilarity

from tqdm import tqdm


class TransEModel(TranslationalModel):
    """Implementation of TransE model detailed in 2013 paper by Bordes et al.. This class inherits from the
    :class:`torchkge.models.interfaces.TranslationalModel` interface. It then has its attributes as well.


    References
    ----------
    * Antoine Bordes, Nicolas Usunier, Alberto Garcia-Duran, Jason Weston, and Oksana Yakhnenko.
      `Translating Embeddings for Modeling Multi-relational Data.
      <https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data>`_
      In Advances in Neural Information Processing Systems 26, pages 2787–2795, 2013.

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.
    dissimilarity_type: String
        Either 'L1' or 'L2'.

    Attributes
    ----------
    relation_embeddings: torch Embedding, shape: (number_relations, ent_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
         normalized.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations, dissimilarity_type):
        try:
            assert dissimilarity_type in ['L1', 'L2', None]
        except AssertionError:
            raise AssertionError("Dissimilarity variable can either be 'L1' or 'L2'.")

        super().__init__(ent_emb_dim, n_entities, n_relations, dissimilarity_type)

        # initialize embeddings
        self.relation_embeddings = init_embedding(self.number_relations, self.ent_emb_dim)

        # normalize parameters
        self.relation_embeddings.weight.data = normalize(self.relation_embeddings.weight.data, p=2, dim=1)
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data, p=2, dim=1)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Score function: opposite of dissimilarities between h+r and t.

        """
        # recover relations embeddings
        rels_emb = self.relation_embeddings(rels_idx)

        # recover, project and normalize entity embeddings
        h_emb = self.recover_project_normalize(heads_idx, rels_idx, normalize_=True)
        t_emb = self.recover_project_normalize(tails_idx, rels_idx, normalize_=True)

        return - self.dissimilarity(h_emb + rels_emb, t_emb)

    def recover_project_normalize(self, ent_idx, rel_idx, normalize_=True):
        """

        Parameters
        ----------
        ent_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of entities
        rel_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of relations
        normalize_: bool
            Whether entities embeddings should be normalized or not.

        Returns
        -------
        projections: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size, ent_emb_dim)
            Embedded entities normalized.

        """
        # recover entity embeddings
        ent_emb = self.entity_embeddings(ent_idx)

        # normalize entity embeddings
        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)

        return ent_emb

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: `torch.Tensor`, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current relations embeddings.

        """
        # recover, project and normalize entity embeddings
        all_idx = arange(0, self.number_entities).long()

        if h_idx.is_cuda:
            all_idx = all_idx.cuda()
        proj_candidates = self.recover_project_normalize(all_idx, None, normalize_=False)

        proj_h_emb = proj_candidates[h_idx]
        proj_t_emb = proj_candidates[t_idx]
        r_emb = self.relation_embeddings(r_idx)

        b_size, emb_dim = proj_h_emb.shape
        proj_candidates = proj_candidates.transpose(0, 1)
        proj_candidates = proj_candidates.view(1, emb_dim, self.number_entities)
        proj_candidates = proj_candidates.expand(b_size, emb_dim, self.number_entities)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb


class TransHModel(TransEModel):
    """Implementation of TransH model detailed in 2014 paper by Wang et al.. This class inherits from the
    :class:`torchkge.models.TranslationalModels.TransEModel` class interpreted as an interface.
    It then has its attributes as well.

    References
    ----------
    * Zhen Wang, Jianwen Zhang, Jianlin Feng, and Zheng Chen.
      `Knowledge Graph Embedding by Translating on Hyperplanes.
      <https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531>`_
      In Twenty-Eighth AAAI Conference on Artificial Intelligence, June 2014.

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    normal_vectors: `torch.Tensor`, shape: (number_relations, ent_emb_dim)
        Normal vectors associated to each relation and used to compute the relation-specific\
        hyperplanes entities are projected on. See paper for more details.

    """

    def __init__(self, ent_emb_dim, n_entities, n_relations):
        super().__init__(ent_emb_dim, n_entities, n_relations, dissimilarity_type='L2')
        self.normal_vectors = Parameter(xavier_uniform_(empty(size=(n_relations, ent_emb_dim))),
                                        requires_grad=True)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Score function: opposite of dissimilarities between h+r and t after projection.

        """
        # recover relations embeddings and normal projection vectors
        relations_embeddings = self.relation_embeddings(rels_idx)
        self.normal_vectors[rels_idx] = normalize(self.normal_vectors[rels_idx], p=2, dim=1)

        # project entities in relation specific hyperplane
        projected_heads = self.recover_project_normalize(heads_idx, rels_idx, normalize_=True)
        projected_tails = self.recover_project_normalize(tails_idx, rels_idx, normalize_=True)

        return - self.dissimilarity(projected_heads + relations_embeddings, projected_tails)

    def recover_project_normalize(self, ent_idx, rel_idx, normalize_=True):
        """Recover entity (either head or tail) embeddings and project on hyperplane defined by\
        provided normal vectors.

        Parameters
        ----------
        ent_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of entities
        rel_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of relations
        normalize_: bool
            Whether entities embeddings should be normalized or not.

        Returns
        -------
        projections: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size, ent_emb_dim)
            Projection of the embedded entities on the hyperplanes defined by the provided normal\
            vectors.

        """
        # recover entity embeddings
        ent_emb = self.entity_embeddings(ent_idx)

        # normalize entity embeddings
        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)

        # project entities into relation space
        normal_vectors = self.normal_vectors[rel_idx]
        normal_component = (ent_emb * normal_vectors).sum(dim=1).view((-1, 1))

        return ent_emb - normal_component * normal_vectors

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.normal_vectors.data = normalize(self.normal_vectors, p=2, dim=1)
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: `torch.Tensor`, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current relations embeddings.

        """
        # recover relations embeddings and normal projection vectors
        r_emb = self.relation_embeddings(r_idx)
        normal_vectors = normalize(self.normal_vectors[r_idx], p=2, dim=1)
        b_size, _ = normal_vectors.shape

        # recover candidates
        candidates = self.recover_candidates(h_idx, b_size)

        # project each candidates with each normal vector
        normal_components = candidates * normal_vectors.view((b_size, self.ent_emb_dim, 1))
        normal_components = normal_components.sum(dim=1).view(b_size, 1, self.number_entities)
        normal_components = normal_components * normal_vectors.view(b_size, self.ent_emb_dim, 1)
        proj_candidates = candidates - normal_components

        assert proj_candidates.shape == (b_size, self.ent_emb_dim, self.number_entities)

        # recover, project and normalize entity embeddings
        proj_h_emb, proj_t_emb = self.projection_helper(h_idx, t_idx, b_size, proj_candidates,
                                                        self.ent_emb_dim)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb


class TransRModel(TranslationalModel):
    """Implementation of TransR model detailed in 2015 paper by Lin et al.. This class inherits from the
    :class:`torchkge.models.interfaces.TranslationalModel` interface. It then has its attributes as well.

    References
    ----------
    * Yankai Lin, Zhiyuan Liu, Maosong Sun, Yang Liu, and Xuan Zhu.
      `Learning Entity and Relation Embeddings for Knowledge Graph Completion.
      <https://www.aaai.org/ocs/index.php/AAAI/AAAI15/paper/view/9571/9523>`_
      In Twenty-Ninth AAAI Conference on Artificial Intelligence, February 2015

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    rel_emb_dim: int
        Dimension of the embedding of relations.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    rel_emb_dim: int
        Dimension of the embedding of relations.
    relation_embeddings: torch.nn.Embedding, shape: (number_relations, rel_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
        normalized.
    projection_matrices: `torch.Tensor`, shape: (number_relations, rel_emb_dim, ent_emb_dim)
        Relation-specific projection matrices. See paper for more details.

    """

    def __init__(self, ent_emb_dim, rel_emb_dim, n_entities, n_relations):

        super().__init__(ent_emb_dim, n_entities, n_relations, dissimilarity_type='L2')

        self.rel_emb_dim = rel_emb_dim
        self.relation_embeddings = init_embedding(self.number_relations, self.rel_emb_dim)

        self.projection_matrices = Parameter(xavier_uniform_(empty(size=(n_relations, rel_emb_dim,
                                                                         ent_emb_dim))),
                                             requires_grad=True)
        self.normalize_parameters()

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Score function: opposite of dissimilarities between h+r and t after projection.

        """
        # recover relations embeddings and normal projection matrices
        relations_embeddings = normalize(self.relation_embeddings(rels_idx), p=2, dim=1)

        # project entities in relation specific hyperplane
        projected_heads = self.recover_project_normalize(heads_idx, rels_idx, normalize_=True)
        projected_tails = self.recover_project_normalize(tails_idx, rels_idx, normalize_=True)

        return - self.dissimilarity(projected_heads + relations_embeddings, projected_tails)

    def recover_project_normalize(self, ent_idx, rel_idx, normalize_=True):
        """Recover entity (either head or tail) embeddings and project on hyperplane defined by\
        provided projection matrices.

        Parameters
        ----------
        ent_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of entities
        rel_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of relations
        normalize_: bool
            Whether entities embeddings should be normalized or not.

        Returns
        -------
        projections: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size, rel_emb_dim)
            Projection of the entities into relation-specific subspaces.
        """
        b_size = ent_idx.shape[0]
        # recover and normalize embeddings
        ent_emb = self.entity_embeddings(ent_idx)
        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)

        # project entities into relation space
        new_shape = (b_size, self.ent_emb_dim, 1)
        projection_matrices = self.projection_matrices[rel_idx]
        projection = matmul(projection_matrices, ent_emb.view(new_shape))

        return normalize(projection.view(b_size, self.rel_emb_dim), p=2, dim=1)

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)
        self.relation_embeddings.weight.data = normalize(self.relation_embeddings.weight.data,
                                                         p=2, dim=1)
        self.projection_matrices.data = normalize(self.projection_matrices.data, p=2, dim=2)

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current relations.

        Returns
        -------
        proj_h_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: `torch.Tensor`, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing all entities projected in each relation spaces (relations
            corresponding to current batch's relations).
        r_emb: `torch.Tensor`, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current relations embeddings.

        """
        # recover relations embeddings and normal projection matrices
        r_emb = normalize(self.relation_embeddings(r_idx), p=2, dim=1)
        projection_matrices = normalize(self.projection_matrices[r_idx], p=2, dim=2)
        b_size, _, _ = projection_matrices.shape

        # recover candidates
        candidates = self.recover_candidates(h_idx, b_size)

        # project each candidates with each projection matrix
        proj_candidates = matmul(projection_matrices, candidates)

        proj_h_emb, proj_t_emb = self.projection_helper(h_idx, t_idx, b_size, candidates,
                                                        self.rel_emb_dim)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb


class TransDModel(TranslationalModel):
    """Implementation of TransD model detailed in 2015 paper by Ji et al.. This class inherits from the
    :class:`torchkge.models.interfaces.TranslationalModel` interface. It then has its attributes as well.

    References
    ----------
    * Guoliang Ji, Shizhu He, Liheng Xu, Kang Liu, and Jun Zhao.
      `Knowledge Graph Embedding via Dynamic Mapping Matrix.
      <https://aclweb.org/anthology/papers/P/P15/P15-1067/>`_
      In Proceedings of the 53rd Annual Meeting of the Association for Computational Linguistics and
      the 7th International Joint Conference on Natural Language Processing (Volume 1: Long Papers)
      pages 687–696, Beijing, China, July 2015. Association for Computational Linguistics.

    Parameters
    ----------
    ent_emb_dim: int
        Dimension of the embedding of entities.
    rel_emb_dim: int
        Dimension of the embedding of relations.
    n_entities: int
        Number of entities in the current data set.
    n_relations: int
        Number of relations in the current data set.

    Attributes
    ----------
    rel_emb_dim: int
        Dimension of the embedding of relations.
    relation_embeddings: torch.nn.Embedding, shape: (number_relations, rel_emb_dim)
        Contains the embeddings of the relations. It is initialized with Xavier uniform and then\
        normalized.
    ent_proj_vects: `torch.Tensor`, shape: (number_entities, ent_emb_dim)
        Entity-specific vector used to build projection matrices. See paper for more details.
    rel_proj_vects: `torch.Tensor`, shape: (number_relations, rel_emb_dim)
        Relation-specific vector used to build projection matrices. See paper for more details.

    """

    def __init__(self, ent_emb_dim, rel_emb_dim, n_entities, n_relations):

        super().__init__(ent_emb_dim, n_entities, n_relations, dissimilarity_type='L2')

        self.rel_emb_dim = rel_emb_dim
        self.relation_embeddings = init_embedding(self.number_relations, self.rel_emb_dim)

        self.ent_proj_vects = Parameter(xavier_uniform_(empty(size=(n_entities, ent_emb_dim))),
                                        requires_grad=True)
        self.rel_proj_vects = Parameter(xavier_uniform_(empty(size=(n_relations, rel_emb_dim))),
                                        requires_grad=True)
        self.normalize_parameters()

        self.evaluated_projections = False
        self.projected_entities = Parameter(empty(size=(self.number_relations,
                                                        self.rel_emb_dim,
                                                        self.number_entities)), requires_grad=False)

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Score function: opposite of dissimilarities between h+r and t after projection.

        """
        self.evaluated_projections = False

        # recover relations projection vectors and relations embeddings
        self.rel_proj_vects[rels_idx] = normalize(self.rel_proj_vects[rels_idx], p=2, dim=1)
        relations_embeddings = normalize(self.relation_embeddings(rels_idx), p=2, dim=1)

        # project
        projected_heads = self.recover_project_normalize(heads_idx, rels_idx, normalize_=True)
        projected_tails = self.recover_project_normalize(tails_idx, rels_idx, normalize_=True)

        return - self.dissimilarity(projected_heads + relations_embeddings, projected_tails)

    def recover_project_normalize(self, ent_idx, rel_idx, normalize_=True):
        """Recover entity (either head or tail) embeddings and project on hyperplane defined by\
        provided normal vectors.

        Parameters
        ----------
        ent_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of entities
        rel_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of relations
        normalize_: bool
            Whether entities embeddings should be normalized or not.

        Returns
        -------
        projections: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size, rel_emb_dim)
            Projection of the entities into relation-specific subspaces.

        """
        b_size = ent_idx.shape[0]

        # recover entities embeddings and projection vectors
        ent_emb = self.entity_embeddings(ent_idx)
        ent_proj = self.ent_proj_vects[ent_idx]

        if normalize_:
            ent_emb = normalize(ent_emb, p=2, dim=1)
            ent_proj = normalize(ent_proj, p=2, dim=1)

        # project entities into relation space
        rel_proj = self.rel_proj_vects[rel_idx]
        proj_mat = matmul(rel_proj.view((b_size, self.rel_emb_dim, 1)),
                          ent_proj.view((b_size, 1, self.ent_emb_dim)))

        if proj_mat.is_cuda:
            proj_mat += eye(n=self.rel_emb_dim, m=self.ent_emb_dim, device='cuda')
        else:
            proj_mat += eye(n=self.rel_emb_dim, m=self.ent_emb_dim)

        projection = matmul(proj_mat, ent_emb.view((b_size, self.ent_emb_dim, 1)))

        return projection.view((b_size, self.rel_emb_dim))

    def normalize_parameters(self):
        """Normalize the parameters of the model using the L2 norm.
        """
        self.entity_embeddings.weight.data = normalize(self.entity_embeddings.weight.data,
                                                       p=2, dim=1)
        self.relation_embeddings.weight.data = normalize(self.relation_embeddings.weight.data,
                                                         p=2, dim=1)
        self.ent_proj_vects.data = normalize(self.ent_proj_vects.data, p=2, dim=1)
        self.rel_proj_vects.data = normalize(self.rel_proj_vects.data, p=2, dim=1)

    def evaluate_projections(self):
        """Project all entities according to each relation.
        """
        if self.evaluated_projections:
            return

        print('Projecting entities in relations spaces.')

        for i in tqdm(range(self.number_entities)):
            ent_proj_vect = self.ent_proj_vects.data[i].view(1, -1)
            rel_proj_vects = self.rel_proj_vects.data.view(self.number_relations,
                                                           self.rel_emb_dim, 1)

            projection_matrices = matmul(rel_proj_vects, ent_proj_vect)

            if projection_matrices.is_cuda:
                id_mat = eye(n=self.rel_emb_dim, m=self.ent_emb_dim, device='cuda')
            else:
                id_mat = eye(n=self.rel_emb_dim, m=self.ent_emb_dim)

            id_mat = id_mat.view(1, self.rel_emb_dim, self.ent_emb_dim)

            projection_matrices += id_mat.expand(self.number_relations, self.rel_emb_dim,
                                                 self.ent_emb_dim)

            empty_cache()

            mask = tensor([i]).long()

            if self.entity_embeddings.weight.is_cuda:
                assert self.projected_entities.is_cuda
                empty_cache()
                mask = mask.cuda()

            entity = self.entity_embeddings(mask)
            projected_entity = matmul(projection_matrices, entity.view(-1)).detach()
            projected_entity = projected_entity.view(self.number_relations, self.rel_emb_dim, 1)
            self.projected_entities[:, :, i] = projected_entity.view(self.number_relations,
                                                                     self.rel_emb_dim)

            del projected_entity

        self.evaluated_projections = True

    def evaluation_helper(self, h_idx, t_idx, r_idx):
        """Project current entities and candidates into relation-specific sub-spaces.

        Parameters
        ----------
        h_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current head entities.
        t_idx: `torch.Tensor`, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current tail entities.
        r_idx: torch Tensor, shape: (b_size), dtype: `torch.long`
            Tensor containing indices of current relations.
        Returns
        -------
        proj_h_emb: torch Tensor, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current head entities projected in relation space.
        proj_t_emb: torch Tensor, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing embeddings of current tail entities projected in relation space.
        proj_candidates: torch Tensor, shape: (b_size, rel_emb_dim, n_entities), dtype: `torch.float`
            Tensor containing all entities projected in each relation spaces (relations\
            corresponding to current batch's relations).
        r_emb: torch Tensor, shape: (b_size, rel_emb_dim), dtype: `torch.float`
            Tensor containing current relations embeddings.

        """
        b_size = h_idx.shape[0]
        if not self.evaluated_projections:
            self.evaluate_projections()

        # recover relations embeddings and projected candidates
        r_emb = normalize(self.relation_embeddings(r_idx), p=2, dim=1)
        proj_candidates = self.projected_entities[r_idx]

        proj_h_emb, proj_t_emb = self.projection_helper(h_idx, t_idx, b_size, proj_candidates,
                                                        self.rel_emb_dim)

        return proj_h_emb, proj_t_emb, proj_candidates, r_emb


class TorusEModel(TransEModel):
    """Implementation of TorusE model detailed in 2018 paper by Ebisu and Ichise. This class inherits from the
    :class:`torchkge.models.TranslationalModels.TransEModel` class interpreted as an interface.
    It then has its attributes as well.

        References
        ----------
        * Takuma Ebisu and Ryutaro Ichise
        `TorusE: Knowledge Graph Embedding on a Lie Group.
        <https://arxiv.org/abs/1711.05435>`_
        In Proceedings of the 32nd AAAI Conference on Artificial Intelligence
        (New Orleans, LA, USA, Feb. 2018), AAAI Press, pp. 1819–1826.

        Parameters
        ----------
        ent_emb_dim: int
            Dimension of the embedding of entities.
        n_entities: int
            Number of entities in the current data set.
        n_relations: int
            Number of relations in the current data set.
        dissimilarity_type: function
            Used to compute dissimilarities (e.g. L1 or L2 dissimilarities).

        Attributes
        ----------
        dissimilarity_type: function
            Used to compute dissimilarities (e.g. L1 or L2 dissimilarities).
        relation_embeddings: torch Embedding, shape: (number_relations, ent_emb_dim)
            Contains the embeddings of the relations. It is initialized with Xavier uniform and\
            then normalized.

        """

    def __init__(self, ent_emb_dim, n_entities, n_relations, dissimilarity_type):

        assert dissimilarity_type in ['L1', 'L2', 'eL2']
        self.dissimilarity_type = dissimilarity_type

        super().__init__(ent_emb_dim, n_entities, n_relations, dissimilarity_type=None)

        self.relation_embeddings = init_embedding(self.number_relations, self.ent_emb_dim)

        if self.dissimilarity_type == 'L1':
            self.dissimilarity = l1_torus_dissimilarity
        if self.dissimilarity_type == 'L2':
            self.dissimilarity = l2_torus_dissimilarity
        if self.dissimilarity_type == 'eL2':
            self.dissimilarity = el2_torus_dissimilarity

        self.normalize_parameters()

    def scoring_function(self, heads_idx, tails_idx, rels_idx):
        """Compute the scoring function for the triplets given as argument.

        Parameters
        ----------
        heads_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's heads
        tails_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's tails.
        rels_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of the current batch's relations.

        Returns
        -------
        score: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size)
            Score function: opposite of dissimilarities between h+r and t.

        """

        # recover relations embeddings
        rels_emb = self.relation_embeddings(rels_idx)

        # recover, project and normalize entity embeddings
        h_emb = self.recover_project_normalize(heads_idx, normalize_=False)
        t_emb = self.recover_project_normalize(tails_idx, normalize_=False)

        if self.dissimilarity_type == 'L1':
            return 2 * self.dissimilarity(h_emb + rels_emb, t_emb)
        if self.dissimilarity_type == 'L2':
            return 4 * self.dissimilarity(h_emb + rels_emb, t_emb)**2
        else:
            assert self.dissimilarity_type == 'eL2'
            return self.dissimilarity(h_emb + rels_emb, t_emb)**2 / 4

    def recover_project_normalize(self, ent_idx, rel_idx=None, normalize_=False):
        """

        Parameters
        ----------
        ent_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of entities
        rel_idx: `torch.Tensor`, dtype: `torch.long`, shape: (batch_size)
            Integer keys of relations
        normalize_: bool
            Whether entities embeddings should be normalized or not.

        Returns
        -------
        projections: `torch.Tensor`, dtype: `torch.float`, shape: (batch_size, ent_emb_dim)
            Embedded entities normalized.

        """
        # recover entity embeddings
        ent_emb = self.entity_embeddings(ent_idx)
        ent_emb.data.frac_()

        return ent_emb

    def normalize_parameters(self):
        """Project embeddings on torus.
        """
        self.entity_embeddings.weight.data.frac_()
        self.relation_embeddings.weight.data.frac_()



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch.nn import Module, MarginRankingLoss, SoftMarginLoss
from torch import ones_like


class MarginLoss(Module):
    """Margin loss as it was defined in `TansE paper
    <https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data>`_
    by Bordes et al. in 2013. This class implements `torch.nn.Module` interface.

    """
    def __init__(self, margin):
        super().__init__()
        self.loss = MarginRankingLoss(margin=margin, reduction='sum')

    def forward(self, output):
        """
        Parameters
        ----------
        output: tuple of two `torch.FloatTensor`
            This tuple of length 2 contains the scores of the golden triplets and the negative triplets as returned by
            the `forward` methods of the models.
        Returns
        -------
        loss: `torch.FloatTensor`, shape: (n_facts, dim)
            Loss of the form :math:`\\max\\{0, \\gamma - f(h,r,t) + f(h',r',t')\\}` where :math:`\\gamma` is the margin
            (defined at initialization), :math:`f(h,r,t)` is the score of a true fact and :math:`f(h',r',t')` is
            the score of the associated negative fact.
        """
        golden_triplets, negative_triplets = output[0], output[1]
        return self.loss(golden_triplets, negative_triplets, target=ones_like(golden_triplets))


class LogisticLoss(Module):
    """Logistic loss as it was defined in `TansE paper
    <https://papers.nips.cc/paper/5071-translating-embeddings-for-modeling-multi-relational-data>`_
    by Bordes et al. in 2013. This class implements `torch.nn.Module` interface.

    """
    def __init__(self):
        super().__init__()
        self.loss = SoftMarginLoss(reduction='sum')

    def forward(self, output):
        """
        Parameters
        ----------
        output: tuple of two `torch.FloatTensor`
            This tuple of length 2 contains the scores of the golden triplets and the negative triplets as returned by
            the `forward` methods of the models.
        Returns
        -------
        loss: `torch.FloatTensor`, shape: (n_facts, dim)
            Loss of the form :math:`\\log(1+ \\exp(\\eta \\times f(h,r,t))` where :math:`f(h,r,t)` is the score of
            the fact and :math:`\\eta` is either 1 or -1 if the fact is true or false.
        """
        golden_triplets, negative_triplets = output[0], output[1]
        targets = ones_like(golden_triplets)
        return self.loss(golden_triplets, targets) + self.loss(negative_triplets, -targets)



from .exceptions import NotYetEvaluatedError
from .exceptions import SizeMismatchError
from .exceptions import WrongDimensionError
from .exceptions import WrongArgumentsError
from .exceptions import NotYetImplementedError
from .exceptions import SanityError
from .exceptions import SplitabilityError



from .KnowledgeGraph import SmallKG
from .KnowledgeGraph import KnowledgeGraph

from .DataLoader import load_fb13
from .DataLoader import load_fb15k
from .DataLoader import load_fb15k237
from .DataLoader import load_wn18
from .DataLoader import load_wikidatasets



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from collections import defaultdict


def get_possible_heads_tails(kg, possible_heads=None, possible_tails=None):
    """Gets for each relation of the knowledge graph the possible heads and possible tails.

    Parameters
    ----------
    kg: `torchkge.data.KnowledgeGraph.KnowledgeGraph`
    possible_heads: dict, optional (default=None)
    possible_tails: dict, optional (default=None)

    Returns
    -------
    possible_heads: dict, optional (default=None)
        keys: relation indices, values: set of possible heads for each relations
    possible_tails: dict, optional (default=None)
        keys: relation indices, values: set of possible tails for each relations

    """

    if possible_heads is None:
        possible_heads = defaultdict(set)
    else:
        assert type(possible_heads) == dict
        possible_heads = defaultdict(set, possible_heads)
    if possible_tails is None:
        possible_tails = defaultdict(set)
    else:
        assert type(possible_tails) == dict
        possible_tails = defaultdict(set, possible_tails)

    for i in range(kg.n_facts):
        possible_heads[kg.relations[i].item()].add(kg.head_idx[i].item())
        possible_tails[kg.relations[i].item()].add(kg.tail_idx[i].item())

    return dict(possible_heads), dict(possible_tails)



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from collections import defaultdict

from torch import cat, eq, int64, long, randperm, tensor, Tensor, zeros_like
from torch.utils.data import Dataset

from torchkge.exceptions import SizeMismatchError, WrongArgumentsError, SanityError
from torchkge.utils import get_dictionaries


class SmallKG(Dataset):
    """Minimalist version of a knowledge graph. Built with tensors of heads, tails and relations.

    """
    def __init__(self, heads, tails, relations):
        assert heads.shape == tails.shape == relations.shape
        self.heads = heads
        self.tails = tails
        self.relations = relations
        self.length = heads.shape[0]

    def __len__(self):
        return self.length

    def __getitem__(self, item):
        return self.heads[item].item(), self.tails[item].item(), self.relations[item].item()


class KnowledgeGraph(Dataset):
    """Knowledge graph representation.

        Parameters
        ----------
        df: pandas.DataFrame
            Data frame containing three columns [from, to, rel].
        ent2ix: dict, optional
            Dictionary mapping entity labels to their integer key.
        rel2ix: dict, optional
            Dictionary mapping relation labels to their integer key.

        Attributes
        ----------
        ent2ix: dict
            Dictionary mapping entity labels to their integer key.
        rel2ix: dict
            Dictionary mapping relation labels to their integer key.
        n_ent: int
            Number of distinct entities in the data set.
        n_rel: int
            Number of distinct entities in the data set.
        n_facts: int
            Number of samples in the data set. A sample is a fact: a triplet (h, r, l).
        head_idx: `torch.Tensor`, dtype = `torch.long`, shape: (n_facts)
            List of the int key of heads for each fact.
        tail_idx: `torch.Tensor`, dtype = `torch.long`, shape: (n_facts)
            List of the int key of tails for each fact.
        relations: `torch.Tensor`, dtype = `torch.long`, shape: (n_facts)
            List of the int key of relations for each fact.

    """

    def __init__(self, df=None, kg=None,
                 ent2ix=None, rel2ix=None, dict_of_heads=None, dict_of_tails=None):

        """
        :param df: `pandas.DataFrame`
        :param kg: dict
            keys should be exhaustively ('heads', 'tails', 'relations')
        :param ent2ix:
        :param rel2ix:
        :param dict_of_heads:
        :param dict_of_tails:
        """

        if df is None:
            if kg is None:
                raise WrongArgumentsError("Please provide at least one argument of `df` and kg`")
            else:
                try:
                    assert (type(kg) == dict) & ('heads' in kg.keys()) & ('tails' in kg.keys()) & \
                           ('relations' in kg.keys())
                except AssertionError:
                    raise WrongArgumentsError("Keys in the `kg` dict should contain `heads`, `tails`, `relations`.")
                try:
                    assert (rel2ix is not None) & (ent2ix is not None)
                except AssertionError:
                    raise WrongArgumentsError("Please provide the two dictionaries ent2ix and rel2ix if building from `kg`.")
        else:
            if kg is not None:
                raise WrongArgumentsError("`df` and kg` arguments should not both provided.")

        if ent2ix is None:
            self.ent2ix = get_dictionaries(df, ent=True)
        else:
            self.ent2ix = ent2ix

        if rel2ix is None:
            self.rel2ix = get_dictionaries(df, ent=False)
        else:
            self.rel2ix = rel2ix

        self.n_ent = max(self.ent2ix.values()) + 1
        self.n_rel = max(self.rel2ix.values()) + 1

        if df is not None:
            # build kg from a pandas dataframe
            self.n_facts = len(df)
            self.head_idx = tensor(df['from'].map(self.ent2ix).values).long()
            self.tail_idx = tensor(df['to'].map(self.ent2ix).values).long()
            self.relations = tensor(df['rel'].map(self.rel2ix).values).long()
        else:
            # build kg from another kg
            self.n_facts = kg['heads'].shape[0]
            self.head_idx = kg['heads']
            self.tail_idx = kg['tails']
            self.relations = kg['relations']

        if dict_of_heads is None or dict_of_tails is None:
            self.dict_of_heads = defaultdict(set)
            self.dict_of_tails = defaultdict(set)
            self.evaluate_dicts()

        else:
            self.dict_of_heads = dict_of_heads
            self.dict_of_tails = dict_of_tails
        try:
            self.sanity_check()
        except AssertionError:
            raise SanityError("Please check the sanity of arguments.")

    def __len__(self):
        return self.n_facts

    def __getitem__(self, item):
        return self.head_idx[item].item(), self.tail_idx[item].item(), self.relations[item].item()

    def sanity_check(self):
        assert (type(self.dict_of_heads) == defaultdict) & (type(self.dict_of_heads) == defaultdict)
        assert (type(self.ent2ix) == dict) & (type(self.rel2ix) == dict)
        assert (len(self.ent2ix) == self.n_ent) & (len(self.rel2ix) == self.n_rel)
        assert (type(self.head_idx) == Tensor) & (type(self.tail_idx) == Tensor) & (type(self.relations) == Tensor)
        assert (self.head_idx.dtype == int64) & (self.tail_idx.dtype == int64) & (self.relations.dtype == int64)
        assert (len(self.head_idx) == len(self.tail_idx) == len(self.relations))

    def split_kg(self, share=0.8, sizes=None, validation=False):
        """Split the knowledge graph into train and test. If `sizes` is provided then it is used to split the
        samples as explained below. If only `share` is provided, the split is done at random but it assures to keep at
        least one fact involving each type of entity and relation in the training subset.

        Parameters
        ----------
        share: float
            Percentage to allocate to train set.
        sizes: tuple
            Tuple of ints of length 2 or 3. If len(sizes) == 2, then the first sizes[0] values of\
            the knowledge graph will be used as training set and the rest as test set.\
            If len(sizes) == 3, the first sizes[0] values of the knowledge graph will be used as\
            training set, the following sizes[1] as validation set and the last sizes[2] as testing\
            set.
        validation: bool
            Indicate if a validation set should be produced along with train and test sets.

        Returns
        -------
        train_kg: `torchkge.data.KnowledgeGraph.KnowledgeGraph`
        val_kg: `torchkge.data.KnowledgeGraph.KnowledgeGraph`, optional
        test_kg: `torchkge.data.KnowledgeGraph.KnowledgeGraph`

        """
        # TODO: assert that all relations in test appear as well in validation (for triplet classification)
        if sizes is not None:
            try:
                if len(sizes) == 3:
                    try:
                        assert (sizes[0] + sizes[1] + sizes[2] == self.n_facts)
                    except AssertionError:
                        raise WrongArgumentsError('Sizes should sum to the number of facts.')
                elif len(sizes) == 2:
                    try:
                        assert (sizes[0] + sizes[1] == self.n_facts)
                    except AssertionError:
                        raise WrongArgumentsError('Sizes should sum to the number of facts.')
                else:
                    raise SizeMismatchError('Tuple `sizes` should be of length 2 or 3.')
            except AssertionError:
                raise SizeMismatchError('Tuple `sizes` should sum up to the number of facts in the '
                                        'knowledge graph.')
        else:
            assert share < 1

        if ((sizes is not None) and (len(sizes) == 3)) or ((sizes is None) and validation):
            # return training, validation and a testing graphs

            if (sizes is None) and validation:
                mask_tr, mask_val, mask_te = self.get_mask(share, validation=True)
            else:
                mask_tr = cat([tensor([1 for _ in range(sizes[0])]),
                               tensor([0 for _ in range(sizes[1] + sizes[2])])]).bool()
                mask_val = cat([tensor([0 for _ in range(sizes[0])]),
                                tensor([1 for _ in range(sizes[1])]),
                                tensor([0 for _ in range(sizes[2])])]).bool()
                mask_te = ~(mask_tr | mask_val)

            return KnowledgeGraph(
                kg={'heads': self.head_idx[mask_tr], 'tails': self.tail_idx[mask_tr],
                    'relations': self.relations[mask_tr]},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails), KnowledgeGraph(
                kg={'heads': self.head_idx[mask_val], 'tails': self.tail_idx[mask_val],
                    'relations': self.relations[mask_val]},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails), KnowledgeGraph(
                kg={'heads': self.head_idx[mask_te], 'tails': self.tail_idx[mask_te],
                    'relations': self.relations[mask_te]},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails)
        else:
            # return training and testing graphs

            assert (((sizes is not None) and len(sizes) == 2) or
                    ((sizes is None) and not validation))
            if sizes is None:
                mask_tr, mask_te = self.get_mask(share, validation=False)
            else:
                mask_tr = cat([tensor([1 for _ in range(sizes[0])]),
                               tensor([0 for _ in range(sizes[1])])]).bool()
                mask_te = ~mask_tr
            return KnowledgeGraph(
                kg={'heads': self.head_idx[mask_tr], 'tails': self.tail_idx[mask_tr],
                    'relations': self.relations[mask_tr]},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails), KnowledgeGraph(
                kg={'heads': self.head_idx[mask_te], 'tails': self.tail_idx[mask_te],
                    'relations': self.relations[mask_te]},
                ent2ix=self.ent2ix, rel2ix=self.rel2ix, dict_of_heads=self.dict_of_heads,
                dict_of_tails=self.dict_of_tails)

    def get_mask(self, share, validation=False):
        """Returns masks to split knowledge graph into train, test and optionally validation sets.
        The mask is first created by dividing samples between subsets based on relation equilibrium.
        Then if any entity is not present in the training subset it is manually added by assigning a share of the sample
        involving the missing entity either as head or tail.

        :param share: float
        :param uniques_e: `torch.Tensor`, dtype: `torch.long`, shape: (m)
        :param uniques_r: `torch.Tensor`, dtype: `torch.long`, shape: (p)
        :param counts_r: `torch.Tensor`, dtype: `torch.long`, shape: (p)
        :param validation: bool
        :return: mask: `torch.Tensor`, dtype: `torch.bool`, shape: (n)
        :return: mask_val: `torch.Tensor`, dtype: `torch.bool`, shape: (n) (optional)
        :return mask_te: `torch.Tensor`, dtype: `torch.bool`, shape: (n)
        """

        uniques_r, counts_r = self.relations.unique(return_counts=True)
        uniques_e, _ = cat((self.head_idx, self.tail_idx)).unique(return_counts=True)

        mask = zeros_like(self.relations).bool()
        if validation:
            mask_val = zeros_like(self.relations).bool()

        # splitting relations among subsets
        for i, r in enumerate(uniques_r):
            rand = randperm(counts_r[i].item())

            sub_mask = eq(self.relations, r).nonzero()[:, 0]  # list of indices k such that relations[k] == r

            assert len(sub_mask) == counts_r[i].item()

            if validation:
                train_size, val_size, test_size = self.get_sizes(counts_r[i], share=share, validation=True)
                mask[sub_mask[rand[:train_size]]] = True
                mask_val[sub_mask[rand[train_size:train_size + val_size]]] = True

            else:
                train_size, test_size = self.get_sizes(counts_r[i], share=share, validation=False)
                mask[sub_mask[rand[:train_size]]] = True

        # adding missing entities to the train set
        u = cat((self.head_idx[mask], self.tail_idx[mask])).unique()
        if len(u) < self.n_ent:
            missing_entities = tensor(list(set(uniques_e.tolist()) - set(u.tolist())), dtype=long)
            for e in missing_entities:
                sub_mask = ((self.head_idx == e) | (self.tail_idx == e)).nonzero()[:, 0]
                rand = randperm(len(sub_mask))
                sizes = self.get_sizes(mask.shape[0], share=share, validation=validation)
                mask[sub_mask[rand[:sizes[0]]]] = True
                if validation:
                    mask_val[sub_mask[rand[:sizes[0]]]] = False

        if validation:
            assert not (mask & mask_val).any().item()
            return mask, mask_val, ~(mask | mask_val)
        else:
            return mask, ~mask

    @staticmethod
    def get_sizes(count, share, validation=False):
        """With `count` samples, returns how many should go to train and test

        """
        if count == 1:
            if validation:
                return 1, 0, 0
            else:
                return 1, 0
        if count == 2:
            if validation:
                return 1, 1, 0
            else:
                return 1, 1

        n_train = int(count * share)
        assert n_train < count
        if n_train == 0:
            n_train += 1

        if not validation:
            return n_train, count - n_train
        else:
            if count - n_train == 1:
                n_train -= 1
                return n_train, 1, 1
            else:
                n_val = int(int(count - n_train) / 2)
                return n_train, n_val, count - n_train - n_val

    def evaluate_dicts(self):
        """Evaluates dicts of possible alternatives to an entity in a fact that still gives a true\
        fact in the entire knowledge graph.

        """
        for i in range(self.n_facts):
            self.dict_of_heads[(self.tail_idx[i].item(),
                                self.relations[i].item())].add(self.head_idx[i].item())
            self.dict_of_tails[(self.head_idx[i].item(),
                                self.relations[i].item())].add(self.tail_idx[i].item())



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from torch import tensor, empty
from torch.nn import Embedding, Parameter
from torch.nn.init import xavier_uniform_


def init_embedding(n_vectors, dim):
    """Create a torch.nn.Embedding object with `n_vectors` samples and `dim` dimensions.
    """
    entity_embeddings = Embedding(n_vectors, dim)
    entity_embeddings.weight = Parameter(xavier_uniform_(empty(size=(n_vectors, dim))), requires_grad=True)
    return entity_embeddings


def get_true_targets(dictionary, e_idx, r_idx, true_idx, i):
    """For a current index `i` of the batch, returns a tensor containing the indices of entities for which the triplet
    is an existing one (i.e. a true one under CWA).

    Parameters
    ----------
    dictionary: default dict
        Dictionary of keys (int, int) and values list of ints giving all possible entities for
        the (entity, relation) pair.
    e_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
        Tensor containing the indices of entities.
    r_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
        Tensor containing the indices of relations.
    true_idx: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
        Tensor containing the true entity for each sample.
    i: int
        Indicates which index of the batch is currently treated.

    Returns
    -------
    true_targets: `torch.Tensor`, shape: (batch_size), dtype: `torch.long`
        Tensor containing the indices of entities such that (e_idx[i], r_idx[i], true_target[any]) is a true fact.

    """
    true_targets = dictionary[e_idx[i].item(), r_idx[i].item()].copy()
    if len(true_targets) == 1:
        return None
    true_targets.remove(true_idx[i].item())
    return tensor(list(true_targets)).long()



# -*- coding: utf-8 -*-
"""
Copyright TorchKGE developers
@author: Armand Boschin <aboschin@enst.fr>
"""

from pandas import DataFrame
from torch import cat


def get_dictionaries(df, ent=True):
    """Build entities or relations dictionaries.

    Parameters
    ----------
    df: `pandas.DataFrame`
        Data frame containing three columns [from, to, rel].
    ent: bool
        if True then ent2ix is returned, if False then rel2ix is returned.

    Returns
    -------
    dict: dictionary
        Either ent2ix or rel2ix.

    """
    if ent:
        tmp = list(set(df['from'].unique()).union(set(df['to'].unique())))
        return {ent: i for i, ent in enumerate(sorted(tmp))}
    else:
        tmp = list(df['rel'].unique())
        return {rel: i for i, rel in enumerate(sorted(tmp))}


def get_tph(t):
    """Get the average number of tail per heads for each relation.

    Parameters
    ----------
    t: `torch.Tensor`, dtype: `torch.long`, shape: (b_size, 3)
        First column contains head indices, second tails and third relations.
    Returns
    -------
    d: dict
        keys: relation indices, values: average number of tail per heads.
    """
    df = DataFrame(t.numpy(), columns=['from', 'to', 'rel'])
    df = df.groupby(['from', 'rel']).count().groupby('rel').mean()
    df.reset_index(inplace=True)
    return {df.loc[i].values[0]: df.loc[i].values[1] for i in df.index}


def get_hpt(t):
    """Get the average number of head per tails for each relation.

    Parameters
    ----------
    t: `torch.Tensor`, dtype: `torch.long`, shape: (b_size, 3)
        First column contains head indices, second tails and third relations.
    Returns
    -------
    d: dict
        keys: relation indices, values: average number of head per tails.
    """
    df = DataFrame(t.numpy(), columns=['from', 'to', 'rel'])
    df = df.groupby(['rel', 'to']).count().groupby('rel').mean()
    df.reset_index(inplace=True)
    return {df.loc[i].values[0]: df.loc[i].values[1] for i in df.index}


def get_bernoulli_probs(kg):
    """Evaluate the Bernoulli probabilities for negative sampling as in the TransH original\
    paper by Wang et al. (2014) https://www.aaai.org/ocs/index.php/AAAI/AAAI14/paper/view/8531.

    Parameters
    ----------
    kg: `torchkge.data.KnowledgeGraph.KnowledgeGraph`

    Returns
    -------
    tph: dict
        keys: relations , values: sampling probabilities as described by Wang et al. in their paper.

    """
    t = cat((kg.head_idx.view(-1, 1), kg.tail_idx.view(-1, 1), kg.relations.view(-1, 1)), dim=1)

    hpt = get_hpt(t)
    tph = get_tph(t)

    assert hpt.keys() == tph.keys()

    for k in tph.keys():
        tph[k] = tph[k] / (tph[k] + hpt[k])

    return tph
